module Filesize
    exposing
        ( defaultSettings
        , format
        , formatWith
        , Settings
        , Units(..)
        )

{-| This library converts a file size in bytes into a human readable string.

Examples:

```
format 1234 == "1.23 kB"
format 238674052 == "238.67 MB"
format 543 == "543 B"
```

You can either use decimal units (also known as base 10 units, these are the
default) or binary (also called base 2 or IEC units).

Supported decimal units:

* 1 byte     (B)
* 1 kilobyte (kB) = 1000 bytes
* 1 megabyte (MB) = 1000 kilobytes
* 1 gigabyte (GB) = 1000 megabytes
* 1 terabyte (TB) = 1000 gigabytes
* 1 petabyte (PB) = 1000 terabytes
* 1 exabyte  (EB) = 1000 petabyte

Larger decimal units (zettabyte (ZB), yottabyte (YB), ...) are not supported.

Supported binary/IEC units:

* 1 byte     (B)
* 1 kibibyte (KiB) = 1024 bytes
* 1 mebibyte (MiB) = 1024 kibibytes
* 1 gibibyte (GiB) = 1024 mebibytes
* 1 tebibyte (TiB) = 1024 gibibytes
* 1 pebibyte (PiB) = 1024 tebibyte

Larger binary units (exbibyte (EiB), zebibyte (ZiB), yobibytej (YiB), ...)) are
not supported.

For decimal/base 10 units, the number of bytes is divided by 10^3 when going to
the next larger unit. For binary/base 2 units, the number of bytes is divided by
2^10 (1024) each time. (For binary units also see
<https://en.wikipedia.org/wiki/Kibibyte>.)

## Usage
@docs format, formatWith, defaultSettings, Settings, Units

-}

import Regex exposing (Regex, HowMany(AtMost))


{-| The two possible unit types, either decimal/base 10 (kb, MB, GB, ...) or
binary/IEC/base 2 (KiB, MiB, GiB, ...), see above.
-}
type Units
    = Base10
    | Base2


{-|
Use a settings record together with `formatWith` to customize the formatting
process. The available options are:

* `units`: use either decimal or binary/IEC units (the default is to use decimal
   units),
* `decimalPlaces`: the number of decimal places (digits after the decimal
   separator, the default is 2) and
* `decimalSeparator`: the decimal separator to use (default ".").
-}
type alias Settings =
    { units : Units
    , decimalPlaces : Int
    , decimalSeparator : String
    }


{-| The default settings. When using `formatWith`, it is recommended to obtain
a settings record with this function and modify the settings to your liking.
-}
defaultSettings : Settings
defaultSettings =
    { units = Base10
    , decimalPlaces = 2
    , decimalSeparator = "."
    }


type alias UnitDefinition =
    { minimumSize : Int
    , abbreviation : String
    }


type alias UnitDefinitionList =
    List UnitDefinition


base10UnitList : UnitDefinitionList
base10UnitList =
    [ { minimumSize = 1, abbreviation = "B" }
    , { minimumSize = 1000, abbreviation = "kB" }
    , { minimumSize = 1000000, abbreviation = "MB" }
    , { minimumSize = 1000000000, abbreviation = "GB" }
    , { minimumSize = 1000000000000, abbreviation = "TB" }
    , { minimumSize = 1000000000000000, abbreviation = "PB" }
    , { minimumSize = 1000000000000000000, abbreviation = "EB" }
      -- , { miniumVersion = 1000000000000000000000, abbreviation = "ZB" }
      -- , { miniumVersion = 1000000000000000000000000, abbreviation = "YB" }
      -- , ...
    ]


base2UnitList : UnitDefinitionList
base2UnitList =
    [ { minimumSize = 1, abbreviation = "B" }
    , { minimumSize = 1024, abbreviation = "KiB" }
    , { minimumSize = 1048576, abbreviation = "MiB" }
    , { minimumSize = 1073741824, abbreviation = "GiB" }
    , { minimumSize = 1099511627776, abbreviation = "TiB" }
    , { minimumSize = 1125899906842624, abbreviation = "PiB" }
      -- , ...
    ]


getUnitDefinitionList : Units -> UnitDefinitionList
getUnitDefinitionList units =
    case units of
        Base10 ->
            base10UnitList

        Base2 ->
            base2UnitList


unknownUnit : UnitDefinition
unknownUnit =
    { minimumSize = 1, abbreviation = "?" }


decimalSeparatorRegex : Regex
decimalSeparatorRegex =
    "." |> Regex.escape |> Regex.regex


{-| Formats the given file size with the default settings.
-}
format : Int -> String
format num =
    formatWith defaultSettings num


{-| Formats the given file size with the given settings.
-}
formatWith : Settings -> Int -> String
formatWith settings num =
    if num == 0 then
        "0 B"
    else
        let
            ( num2, negativePrefix ) =
                if num < 0 then
                    ( num |> negate, "-" )
                else
                    ( num, "" )

            unitDefinitionList =
                getUnitDefinitionList settings.units

            unitDefinition =
                unitDefinitionList
                    |> List.filter (\unitDef -> num2 >= unitDef.minimumSize)
                    |> List.reverse
                    |> List.head
                    |> Maybe.withDefault unknownUnit

            formattedNumber =
                (toFloat num2)
                    / (toFloat unitDefinition.minimumSize)
                    |> roundToDecimalPlaces settings
        in
            negativePrefix ++ formattedNumber ++ " " ++ unitDefinition.abbreviation


roundToDecimalPlaces : Settings -> Float -> String
roundToDecimalPlaces settings num =
    let
        decimalPlaces =
            if settings.decimalPlaces >= 0 then
                settings.decimalPlaces
            else
                defaultSettings.decimalPlaces

        factor =
            (10 ^ (toFloat decimalPlaces))

        -- Actually, using round instead of floor would be preferable but we
        -- never want to round from 999.999 to 1000.
        scaledValue =
            num * factor |> floor |> toFloat

        rounded =
            toString (scaledValue / factor)
    in
        if (settings.decimalSeparator == ".") then
            rounded
        else
            Regex.replace
                (AtMost 1)
                decimalSeparatorRegex
                (\_ -> settings.decimalSeparator)
                rounded
